@@ -1,47 +1,45 @@
-/*  Tarlz - Archiver with multimember lzip compression
-    Copyright (C) 2013-2019 Antonio Diaz Diaz.
+/* Tarlz - Archiver with multimember lzip compression
+   Copyright (C) 2013-2023 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
 
 #include <algorithm>
 #include <cerrno>
-#include <climits>
 #include <cstdio>
-#include <cstdlib>
-#include <cstring>
 #include <queue>
-#include <string>
-#include <vector>
 #include <pthread.h>
-#include <stdint.h>
+#include <stdint.h>	// for lzlib.h
 #include <unistd.h>
 #include <sys/stat.h>
 #include <ftw.h>
 #include <lzlib.h>
 
-#include "arg_parser.h"
 #include "tarlz.h"
+#include "arg_parser.h"
+#include "common_mutex.h"
+#include "create.h"
 
 
 namespace {
 
+const Cl_options * gcl_opts = 0;	// local vars needed by add_member_lz
 enum { max_packet_size = 1 << 20 };
 class Packet_courier;
-Packet_courier * courierp = 0;		// local vars needed by add_member_lz
+Packet_courier * courierp = 0;
 unsigned long long partial_data_size = 0;	// size of current block
 
 
@@ -83,15 +81,15 @@ public:
 
 struct Ipacket			// filename, file size and headers
   {
-  const unsigned long long file_size;
+  const long long file_size;
   const std::string filename;	// filename.empty() means end of lzip member
   const Extended * const extended;
   const uint8_t * const header;
 
   Ipacket() : file_size( 0 ), extended( 0 ), header( 0 ) {}
-  Ipacket( const char * const name, const unsigned long long s,
+  Ipacket( const char * const name, const long long fs,
            const Extended * const ext, const uint8_t * const head )
-    : file_size( s ), filename( name ), extended( ext ), header( head ) {}
+    : file_size( fs ), filename( name ), extended( ext ), header( head ) {}
   };
 
 struct Opacket		// compressed data to be written to the archive
@@ -156,7 +154,7 @@ public:
      If filename.empty() (end of lzip member token), move to next queue. */
   void receive_packet( const Ipacket * const ipacket )
     {
-    if( ipacket->filename.size() )
+    if( !ipacket->filename.empty() )
       slot_tally.get_slot();		// wait for a free slot
     xlock( &imutex );
     ipacket_queues[receive_worker_id].push( ipacket );
@@ -184,7 +182,7 @@ public:
       }
     xunlock( &imutex );
     if( ipacket )
-      { if( ipacket->filename.size() ) slot_tally.leave_slot(); }
+      { if( !ipacket->filename.empty() ) slot_tally.leave_slot(); }
     else
       {
       // notify muxer when last worker exits
@@ -260,8 +258,8 @@ public:
 int add_member_lz( const char * const filename, const struct stat *,
                    const int flag, struct FTW * )
   {
-  if( Exclude::excluded( filename ) ) return 0;		// skip excluded
-  unsigned long long file_size = 0;
+  if( Exclude::excluded( filename ) ) return 0;		// skip excluded files
+  long long file_size;
   // metadata for extended records
   Extended * const extended = new( std::nothrow ) Extended;
   uint8_t * const header = extended ? new( std::nothrow ) Tar_header : 0;
@@ -269,14 +267,19 @@ int add_member_lz( const char * const filename, const struct stat *,
     { show_error( mem_msg ); if( extended ) delete extended; return 1; }
   if( !fill_headers( filename, *extended, header, file_size, flag ) )
     { delete[] header; delete extended; return 0; }
+  print_removed_prefix( extended->removed_prefix );
 
-  if( solidity == bsolid &&
-      block_is_full( *extended, file_size, partial_data_size ) )
-    courierp->receive_packet( new Ipacket );		// end of group
-
+  if( gcl_opts->solidity == bsolid )
+    {
+    const int ebsize = extended->full_size();
+    if( ebsize < 0 ) { show_error( extended->full_size_error() ); return 1; }
+    if( block_is_full( ebsize, file_size, gcl_opts->data_size,
+                       partial_data_size ) )
+      courierp->receive_packet( new Ipacket );		// end of group
+    }
   courierp->receive_packet( new Ipacket( filename, file_size, extended, header ) );
 
-  if( solidity == no_solid )			// one tar member per group
+  if( gcl_opts->solidity == no_solid )		// one tar member per group
     courierp->receive_packet( new Ipacket );
   if( verbosity >= 1 ) std::fprintf( stderr, "%s\n", filename );
   return 0;
@@ -285,58 +288,56 @@ int add_member_lz( const char * const filename, const struct stat *,
 
 struct Grouper_arg
   {
+  const Cl_options * cl_opts;
   Packet_courier * courier;
-  const Arg_parser * parser;
-  bool dereference;
   };
 
 
 /* Package metadata of the files to be archived and pass them to the
-   courier for distribution to workers. */
+   courier for distribution to workers.
+*/
 extern "C" void * grouper( void * arg )
   {
   const Grouper_arg & tmp = *(const Grouper_arg *)arg;
+  const Cl_options & cl_opts = *tmp.cl_opts;
   Packet_courier & courier = *tmp.courier;
-  const Arg_parser & parser = *tmp.parser;
-  const bool dereference = tmp.dereference;
 
-  for( int i = 0; i < parser.arguments(); ++i )		// parse command line
+  for( int i = 0; i < cl_opts.parser.arguments(); ++i )	// parse command line
     {
-    const int code = parser.code( i );
-    const std::string & arg = parser.argument( i );
+    const int code = cl_opts.parser.code( i );
+    const std::string & arg = cl_opts.parser.argument( i );
     const char * filename = arg.c_str();
     if( code == 'C' && chdir( filename ) != 0 )
-      { show_file_error( filename, "Error changing working directory", errno );
-        cleanup_and_fail(); }
+      { show_file_error( filename, chdir_msg, errno ); exit_fail_mt(); }
     if( code ) continue;				// skip options
-    if( parser.argument( i ).empty() ) continue;	// skip empty names
+    if( cl_opts.parser.argument( i ).empty() ) continue; // skip empty names
     std::string deslashed;		// arg without trailing slashes
     unsigned len = arg.size();
     while( len > 1 && arg[len-1] == '/' ) --len;
     if( len < arg.size() )
       { deslashed.assign( arg, 0, len ); filename = deslashed.c_str(); }
-    if( Exclude::excluded( filename ) ) continue;	// skip excluded
+    if( Exclude::excluded( filename ) ) continue;	// skip excluded files
     struct stat st;
     if( lstat( filename, &st ) != 0 )	// filename from command line
-      { show_file_error( filename, "Can't stat input file", errno );
-        set_error_status( 1 ); }
+      { show_file_error( filename, cant_stat, errno ); set_error_status( 1 ); }
     else if( nftw( filename, add_member_lz, 16,
-                   dereference ? 0 : FTW_PHYS ) != 0 )
-      cleanup_and_fail();			// write error or oom
-    else if( solidity == dsolid )		// end of group
+                   cl_opts.dereference ? 0 : FTW_PHYS ) != 0 )
+      exit_fail_mt();			// write error or OOM
+    else if( cl_opts.solidity == dsolid )	// end of group
       courier.receive_packet( new Ipacket );
     }
 
-  if( solidity == bsolid && partial_data_size )		// finish last block
+  if( cl_opts.solidity == bsolid && partial_data_size )	// finish last block
     { partial_data_size = 0; courierp->receive_packet( new Ipacket ); }
   courier.finish();			// no more packets to send
   return 0;
   }
 
 
-/* Writes ibuf to encoder. To minimize dictionary size, it does not read
-   from encoder until encoder's input buffer is full or finish is true.
-   Sends opacket to courier and allocates new obuf each time obuf is full. */
+/* Write ibuf to encoder. To minimize dictionary size, do not read from
+   encoder until encoder's input buffer is full or finish is true.
+   Send opacket to courier and allocate new obuf each time obuf is full.
+*/
 void loop_encode( const uint8_t * const ibuf, const int isize,
                   uint8_t * & obuf, int & opos, Packet_courier & courier,
                   LZ_Encoder * const encoder, const int worker_id,
@@ -362,7 +363,7 @@ void loop_encode( const uint8_t * const ibuf, const int isize,
       if( verbosity >= 0 )
         std::fprintf( stderr, "LZ_compress_read error: %s\n",
                       LZ_strerror( LZ_compress_errno( encoder ) ) );
-      cleanup_and_fail();
+      exit_fail_mt();
       }
     opos += rd;
     // obuf is full or last opacket in lzip member
@@ -372,11 +373,11 @@ void loop_encode( const uint8_t * const ibuf, const int isize,
         internal_error( "opacket size exceeded in worker." );
       courier.collect_packet( new Opacket( obuf, opos ), worker_id );
       opos = 0; obuf = new( std::nothrow ) uint8_t[max_packet_size];
-      if( !obuf ) { show_error( mem_msg2 ); cleanup_and_fail(); }
+      if( !obuf ) { show_error( mem_msg2 ); exit_fail_mt(); }
       if( LZ_compress_finished( encoder ) == 1 )
         {
         if( LZ_compress_restart_member( encoder, LLONG_MAX ) >= 0 ) break;
-        show_error( "LZ_compress_restart_member failed." ); cleanup_and_fail();
+        show_error( "LZ_compress_restart_member failed." ); exit_fail_mt();
         }
       }
     }
@@ -395,7 +396,8 @@ struct Worker_arg
 
 
 /* Get ipackets from courier, compress headers and file data, and give the
-   opackets produced to courier. */
+   opackets produced to courier.
+*/
 extern "C" void * cworker( void * arg )
   {
   const Worker_arg & tmp = *(const Worker_arg *)arg;
@@ -407,7 +409,7 @@ extern "C" void * cworker( void * arg )
   LZ_Encoder * encoder = 0;
   uint8_t * data = 0;
   Resizable_buffer rbuf;			// extended header + data
-  if( !rbuf.size() ) { show_error( mem_msg2 ); cleanup_and_fail(); }
+  if( !rbuf.size() ) { show_error( mem_msg2 ); exit_fail_mt(); }
 
   int opos = 0;
   bool flushed = true;		// avoid producing empty lzip members
@@ -423,8 +425,8 @@ extern "C" void * cworker( void * arg )
       flushed = true; delete ipacket; continue;
       }
 
-    const int infd =
-      ipacket->file_size ? open_instream( ipacket->filename.c_str() ) : -1;
+    const char * const filename = ipacket->filename.c_str();
+    const int infd = ipacket->file_size ? open_instream( filename ) : -1;
     if( ipacket->file_size && infd < 0 )	// can't read file data
       { delete[] ipacket->header; delete ipacket->extended; delete ipacket;
         set_error_status( 1 ); continue; }	// skip file
@@ -440,22 +442,15 @@ extern "C" void * cworker( void * arg )
           show_error( mem_msg2 );
         else
           internal_error( "invalid argument to encoder." );
-        cleanup_and_fail();
+        exit_fail_mt();
         }
       }
 
-    if( !ipacket->extended->empty() )		// compress extended block
-      {
-      const long long ebsize = ipacket->extended->format_block( rbuf );
-      if( ebsize < 0 )
-        { show_error( "Error formatting extended records." ); cleanup_and_fail(); }
-      /* Limit the size of the extended block to INT_MAX - 1 so that it can
-         be fed to lzlib as one buffer. */
-      if( ebsize >= INT_MAX )
-        { show_error( "Extended records size >= INT_MAX." ); cleanup_and_fail(); }
-      loop_encode( (const uint8_t *)rbuf(), ebsize, data, opos, courier,
-                   encoder, worker_id );
-      }
+    const int ebsize = ipacket->extended->format_block( rbuf );	// may be 0
+    if( ebsize < 0 )
+      { show_error( ipacket->extended->full_size_error() ); exit_fail_mt(); }
+    if( ebsize > 0 )				// compress extended block
+      loop_encode( rbuf.u8(), ebsize, data, opos, courier, encoder, worker_id );
     // compress ustar header
     loop_encode( ipacket->header, header_size, data, opos, courier,
                  encoder, worker_id );
@@ -463,20 +458,18 @@ extern "C" void * cworker( void * arg )
 
     if( ipacket->file_size )
       {
-      enum { bufsize = 32 * header_size };
+      const long long bufsize = 32 * header_size;
       uint8_t buf[bufsize];
-      unsigned long long rest = ipacket->file_size;
+      long long rest = ipacket->file_size;
       while( rest > 0 )
         {
-        int size = std::min( rest, (unsigned long long)bufsize );
+        int size = std::min( rest, bufsize );
         const int rd = readblock( infd, buf, size );
         rest -= rd;
         if( rd != size )
           {
-          if( verbosity >= 0 )
-            std::fprintf( stderr, "File '%s' ends unexpectedly at pos %llu\n",
-                          ipacket->filename.c_str(), ipacket->file_size - rest );
-          close( infd ); cleanup_and_fail();
+          show_atpos_error( filename, ipacket->file_size - rest, false );
+          close( infd ); exit_fail_mt();
           }
         if( rest == 0 )				// last read
           {
@@ -489,20 +482,23 @@ extern "C" void * cworker( void * arg )
         loop_encode( buf, size, data, opos, courier, encoder, worker_id );
         }
       if( close( infd ) != 0 )
-        { show_file_error( ipacket->filename.c_str(), "Error closing file", errno );
-          cleanup_and_fail(); }
+        { show_file_error( filename, eclosf_msg, errno ); exit_fail_mt(); }
       }
+    if( gcl_opts->warn_newer && archive_attrs.is_newer( filename ) )
+      { show_file_error( filename, "File is newer than the archive." );
+        set_error_status( 1 ); }
     delete ipacket;
     }
   if( data ) delete[] data;
   if( encoder && LZ_compress_close( encoder ) < 0 )
-    { show_error( "LZ_compress_close failed." ); cleanup_and_fail(); }
+    { show_error( "LZ_compress_close failed." ); exit_fail_mt(); }
   return 0;
   }
 
 
 /* Get from courier the processed and sorted packets, and write
-   their contents to the output archive. */
+   their contents to the output archive.
+*/
 void muxer( Packet_courier & courier, const int outfd )
   {
   while( true )
@@ -511,7 +507,7 @@ void muxer( Packet_courier & courier, const int outfd )
     if( !opacket ) break;	// queue is empty. all workers exited
 
     if( !writeblock_wrapper( outfd, opacket->data, opacket->size ) )
-      cleanup_and_fail();
+      exit_fail_mt();
     delete[] opacket->data;
     delete opacket;
     }
@@ -521,34 +517,35 @@ void muxer( Packet_courier & courier, const int outfd )
 
 
 // init the courier, then start the grouper and the workers and call the muxer
-int encode_lz( const Arg_parser & parser, const int dictionary_size,
-               const int match_len_limit, const int num_workers,
-               const int outfd, const int out_slots, const int debug_level,
-               const bool dereference )
+int encode_lz( const Cl_options & cl_opts, const char * const archive_namep,
+               const int outfd )
   {
   const int in_slots = 65536;		// max small files (<=512B) in 64 MiB
+  const int num_workers = cl_opts.num_workers;
   const int total_in_slots = ( INT_MAX / num_workers >= in_slots ) ?
                              num_workers * in_slots : INT_MAX;
+  const int dictionary_size = option_mapping[cl_opts.level].dictionary_size;
+  const int match_len_limit = option_mapping[cl_opts.level].match_len_limit;
+  gcl_opts = &cl_opts;
 
   /* If an error happens after any threads have been started, exit must be
      called before courier goes out of scope. */
-  Packet_courier courier( num_workers, total_in_slots, out_slots );
+  Packet_courier courier( num_workers, total_in_slots, cl_opts.out_slots );
   courierp = &courier;			// needed by add_member_lz
 
   Grouper_arg grouper_arg;
+  grouper_arg.cl_opts = &cl_opts;
   grouper_arg.courier = &courier;
-  grouper_arg.parser = &parser;
-  grouper_arg.dereference = dereference;
 
   pthread_t grouper_thread;
   int errcode = pthread_create( &grouper_thread, 0, grouper, &grouper_arg );
   if( errcode )
-    { show_error( "Can't create grouper thread", errcode ); cleanup_and_fail(); }
+    { show_error( "Can't create grouper thread", errcode ); return 1; }
 
   Worker_arg * worker_args = new( std::nothrow ) Worker_arg[num_workers];
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
   if( !worker_args || !worker_threads )
-    { show_error( mem_msg ); cleanup_and_fail(); }
+    { show_error( mem_msg ); exit_fail_mt(); }
   for( int i = 0; i < num_workers; ++i )
     {
     worker_args[i].courier = &courier;
@@ -557,7 +554,7 @@ int encode_lz( const Arg_parser & parser, const int dictionary_size,
     worker_args[i].worker_id = i;
     errcode = pthread_create( &worker_threads[i], 0, cworker, &worker_args[i] );
     if( errcode )
-      { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
+      { show_error( "Can't create worker threads", errcode ); exit_fail_mt(); }
     }
 
   muxer( courier, outfd );
@@ -566,22 +563,22 @@ int encode_lz( const Arg_parser & parser, const int dictionary_size,
     {
     errcode = pthread_join( worker_threads[i], 0 );
     if( errcode )
-      { show_error( "Can't join worker threads", errcode ); cleanup_and_fail(); }
+      { show_error( "Can't join worker threads", errcode ); exit_fail_mt(); }
     }
   delete[] worker_threads;
   delete[] worker_args;
 
   errcode = pthread_join( grouper_thread, 0 );
   if( errcode )
-    { show_error( "Can't join grouper thread", errcode ); cleanup_and_fail(); }
+    { show_error( "Can't join grouper thread", errcode ); exit_fail_mt(); }
 
   // write End-Of-Archive records
-  int retval = !write_eof_records( outfd, true );
+  int retval = !write_eoa_records( outfd, true );
 
-  if( close( outfd ) != 0 && !retval )
-    { show_error( "Error closing archive", errno ); retval = 1; }
+  if( close( outfd ) != 0 && retval == 0 )
+    { show_file_error( archive_namep, eclosa_msg, errno ); retval = 1; }
 
-  if( debug_level & 1 )
+  if( cl_opts.debug_level & 1 )
     std::fprintf( stderr,
       "any worker tried to consume from grouper %8u times\n"
       "any worker had to wait                   %8u times\n"
@@ -592,6 +589,6 @@ int encode_lz( const Arg_parser & parser, const int dictionary_size,
       courier.ocheck_counter,
       courier.owait_counter );
 
-  if( !courier.finished() ) internal_error( "courier not finished." );
+  if( !courier.finished() ) internal_error( conofin_msg );
   return final_exit_status( retval );
   }
